#ifndef M68000H
#define M68000H

/*
  Header 68000 emulation
  Petter Schau 1996
*/

#define M68000  0
#define M68010  1
#define M68020  2
#define M68030  3
#define M68EC30 4
#define M68040  5
#define M68EC40 6
#define M68060  7
#define M68EC60 8

typedef void (*cpuStackFrameFunc)(ULO);
extern void cpuGroup1(ULO);
extern void cpuGroup2(ULO);
extern void cpuFrame0(ULO);
extern void cpuFrame1(ULO);
extern void cpuFrame2(ULO);
extern void cpuFrame3(ULO);
extern void cpuFrame4(ULO);
extern void cpuFrame7(ULO);
extern void cpuFrame8(ULO);
extern void cpuFrame9(ULO);
extern void cpuFrameA(ULO);
extern void cpuFrameB(ULO);


extern void m68000start(void);   /* Code alignment label */

extern ULO cpu_next_instruction(void);

#ifdef PREFETCH
#pragma aux prefetch_fill parm [edx];
extern void prefetch_fill(ULO address);
extern ULO prefetch_word;
extern ULO prefetch_pos;
extern ULO prefetch_size;
#endif


#pragma aux set_pc parm [edx];
extern void set_pc(ULO address);

#pragma aux get_pc parm [edx] value [edx];
extern ULO get_pc(ULO address);

extern ULO config_cpu_type;
extern ULO config_cpu_speed;
extern ULO cpu_next, irq_next;

extern ULO exceptionstack;
extern ULO exceptionbackdooraddress;
extern void cpuinit(void);
extern void prepare_exception(void);
extern void next_instruction(void);
extern void privilegie_violation(void);
extern void updateSR(void);
extern ULO pc;
extern ULO thiscycle;
extern ULO t[65536][8];
extern ULO a[8],d[8];
extern ULO sr,usp,ssp;
extern ULO cpuspeed;
extern ULO cpustopflag;
extern ULO interruptflag;
extern ULO statustab[4096];
extern ULO ccrlo_new, ccrhi_new;

#ifdef PC_PTR
extern ULO pcbaseadr;
#endif

extern ULO cpu_dis_tab[65536];

#ifdef PROF68
void dump_profile(void);
void init_profile(void);
#endif

extern const UBY allmodes[16];
extern const UBY data[16];
extern const UBY memory[16];
extern const UBY control[16];
extern const UBY alterable[16];
extern UBY tarb[12];
extern UBY tarw[12];
extern UBY tarl[12];
extern eareadfunc arb[12],arw[12],arl[12],parb[12],parw[12],parl[12];
extern eawritefunc awb[12],aww[12],awl[12];
extern eacalcfunc eac[12];
extern eareadfunc arb68000[12];
extern eareadfunc arw68000[12];
extern eareadfunc arl68000[12];
extern eareadfunc parb68000[12];
extern eareadfunc parw68000[12];
extern eareadfunc parl68000[12];
extern eawritefunc awb68000[12];
extern eawritefunc aww68000[12];
extern eawritefunc awl68000[12];
extern eacalcfunc eac68000[12];
extern const UBY tarb68000[12];
extern const UBY tarw68000[12];
extern const UBY tarl68000[12];
extern UBY ccctt[16];
extern UBY ccdtt[16];
extern UBY ccett[16];
extern UBY ccftt[16];
extern UBY hex2bcd[256];
extern UBY bcd2hex[256];

void set_up_interrupt(void);

#endif
